<?php

namespace com\fwdekker\deathnotifier\trackings;

use com\fwdekker\deathnotifier\ArticleType;
use com\fwdekker\deathnotifier\Database;
use com\fwdekker\deathnotifier\mailer\Mailer;
use com\fwdekker\deathnotifier\mailer\NotifyArticleDeletedEmail;
use com\fwdekker\deathnotifier\mailer\NotifyArticleUndeletedEmail;
use com\fwdekker\deathnotifier\mailer\NotifyStatusChangedEmail;
use com\fwdekker\deathnotifier\Mediawiki;
use com\fwdekker\deathnotifier\Response;
use Exception;
use Monolog\Logger;
use PDO;


/**
 * Manages interaction with the database in the context of trackings.
 */
class TrackingManager
{
    /**
     * The minimum length of a Wikipedia article title.
     */
    public const MIN_TITLE_LENGTH = 1;
    /**
     * The maximum length of a Wikipedia article title.
     */
    public const MAX_TITLE_LENGTH = 255;

    /**
     * @var Logger The logger to log with.
     */
    private Logger $logger;
    /**
     * @var PDO The database connection to interact with.
     */
    private PDO $conn;
    /**
     * @var Mailer The mailer to send emails with.
     */
    private Mailer $mailer;
    /**
     * @var Mediawiki The Mediawiki instance to use for interacting with Wikipedia.
     */
    private Mediawiki $mediawiki;


    /**
     * Constructs a new tracking manager.
     *
     * @param Logger $logger the logger to log with
     * @param PDO $conn the database connection to interact with
     * @param Mailer $mailer the mailer to send emails with
     * @param Mediawiki $mediawiki the Mediawiki instance to use for interacting with Wikipedia
     */
    public function __construct(Logger $logger, PDO $conn, Mailer $mailer, Mediawiki $mediawiki)
    {
        $this->logger = $logger;
        $this->conn = $conn;
        $this->mailer = $mailer;
        $this->mediawiki = $mediawiki;
    }


    /**
     * Populates the database with the necessary structures for trackings.
     *
     * @return void
     */
    public function install(): void
    {
        $this->conn->exec("CREATE TABLE trackings(user_uuid TEXT NOT NULL,
                                                  person_name TEXT NOT NULL,
                                                  PRIMARY KEY (user_uuid, person_name),
                                                  FOREIGN KEY (user_uuid) REFERENCES users (uuid)
                                                      ON DELETE CASCADE
                                                      ON UPDATE CASCADE,
                                                  FOREIGN KEY (person_name) REFERENCES people (name)
                                                      ON DELETE CASCADE
                                                      ON UPDATE CASCADE);");
        // TODO: Add column for date since tracking
        $this->conn->exec("CREATE TABLE people(name TEXT NOT NULL UNIQUE PRIMARY KEY,
                                               status TEXT NOT NULL DEFAULT(''),
                                               is_deleted INT NOT NULL DEFAULT(0));");
        $this->conn->exec("CREATE TRIGGER people_cull_orphans
                           AFTER DELETE ON trackings
                           FOR EACH ROW
                           WHEN (SELECT COUNT(*) FROM trackings WHERE person_name=OLD.person_name)=0
                           BEGIN
                               DELETE FROM people WHERE name=OLD.person_name;
                           END;");
    }


    /**
     * Adds a tracking to the database.
     *
     * @param string $user_uuid the user to whom the tracking belongs
     * @param string $person_name the name of the person to track
     * @return Response a satisfied `Response` with payload `null` if the tracking was added, or an unsatisfied
     * `Response` otherwise
     */
    public function add_tracking(string $user_uuid, string $person_name): Response
    {
        try {
            $info = $this->mediawiki->query_page_info([$person_name]);
        } catch (Exception $exception) {
            $this->logger->error("Failed to query page info.", ["cause" => $exception, "name" => $person_name]);
            return Response::unsatisfied("Could not reach Wikipedia. Maybe the website is down?");
        }

        $normalized_name = $info->redirects[$person_name];
        $status = $info->results[$normalized_name]["status"];
        $type = $info->results[$normalized_name]["type"];

        if (in_array($normalized_name, $info->missing))
            return Response::unsatisfied(
                "Wikipedia does not have an article about " .
                "<b><a href='https://en.wikipedia.org/wiki/Special:Search?search=" .
                rawurlencode($normalized_name) . "'>" . htmlentities($person_name) . "</a></b>. " .
                "Maybe you need to capitalise the surname?",
                "person_name"
            );
        if ($type === ArticleType::Disambiguation)
            return Response::unsatisfied(
                "<b><a href='https://en.wikipedia.org/wiki/" . rawurlencode($normalized_name) . "'>" .
                htmlentities($normalized_name) . "</a></b> refers to multiple articles. " .
                "<a href='https://en.wikipedia.org/wiki/" . rawurlencode($normalized_name) . "'>Check Wikipedia</a> " .
                "to see if your article is listed.",
                "person_name"
            );
        if ($type === ArticleType::Other)
            return Response::unsatisfied(
                "The Wikipedia article about " .
                "<b><a href='https://en.wikipedia.org/wiki/" . rawurlencode($normalized_name) . "'>" .
                htmlentities($normalized_name) . "</a></b> is not about a person.",
                "person_name"
            );
        if ($status === null)
            return Response::unsatisfied("Failed to understand the person's status. This is a bug. Try again maybe?");

        // Insert person and tracking
        return Database::transaction(
            $this->conn,
            function () use ($user_uuid, $person_name, $normalized_name, $status) {
                $stmt = $this->conn->prepare("SELECT EXISTS(SELECT 1
                                              FROM trackings
                                              WHERE user_uuid=:uuid AND person_name=:name);");
                $stmt->bindValue(":uuid", $user_uuid);
                $stmt->bindValue(":name", $normalized_name);
                $stmt->execute();
                if ($stmt->fetch()[0] === 1)
                    return Response::unsatisfied(
                        "You are already tracking <b>" . htmlentities($normalized_name) . "</b>.",
                        "person_name"
                    );

                $stmt = $this->conn->prepare("INSERT OR IGNORE INTO people (name) VALUES (:name);");
                $stmt->bindValue(":name", $normalized_name);
                $stmt->execute();

                $stmt = $this->conn->prepare("UPDATE people SET status=:status WHERE name=:name;");
                $stmt->bindValue(":name", $normalized_name);
                $stmt->bindValue(":status", $status->value);
                $stmt->execute();

                $stmt = $this->conn->prepare("INSERT OR IGNORE INTO trackings (user_uuid, person_name)
                                              VALUES (:user_uuid, :person_name);");
                $stmt->bindValue(":user_uuid", $user_uuid);
                $stmt->bindValue(":person_name", $normalized_name);
                $stmt->execute();

                return Response::satisfied([
                    "name" => $normalized_name,
                    "input" => $person_name,
                    "renamed" => strtolower($person_name) !== strtolower($normalized_name)
                ]);
            });
    }

    /**
     * Removes a tracking from the database.
     *
     * @param string $user_uuid the user to whom the tracking belongs
     * @param string $person_name the name of the tracked person to remove
     * @return Response a satisfied `Response` with payload `null`
     */
    public function remove_tracking(string $user_uuid, string $person_name): Response
    {
        $stmt = $this->conn->prepare("DELETE FROM trackings
                                      WHERE user_uuid=:user_uuid AND person_name=:person_name;");
        $stmt->bindValue(":user_uuid", $user_uuid);
        $stmt->bindValue(":person_name", $person_name);
        $stmt->execute();

        return Response::satisfied();
    }

    /**
     * Lists all trackings of the indicated user.
     *
     * @param string $user_uuid the user to return the trackings of
     * @return Response a satisfied `Response` with the trackings as its payload
     */
    public function list_trackings(string $user_uuid): Response
    {
        $stmt = $this->conn->prepare("SELECT people.name, people.status, people.is_deleted
                                      FROM trackings
                                      INNER JOIN people
                                          ON trackings.user_uuid=:user_uuid
                                          AND trackings.person_name=people.name;");
        $stmt->bindValue(":user_uuid", $user_uuid);
        $stmt->execute();
        $results = $stmt->fetchAll(PDO::FETCH_ASSOC);

        return Response::satisfied($results);
    }

    /**
     * Returns the email addresses of all users who should be notified of events relating to the given person.
     *
     * @param string $person_name the person to receive subscribed email addresses for
     * @return string[] the email addresses of all users who should be notified of events relating to the given person
     */
    private function get_recipients(string $person_name): array
    {
        $get_recipients = $this->conn->prepare("SELECT users.email
                                                FROM users
                                                LEFT JOIN trackings
                                                    WHERE trackings.person_name=:person_name
                                                    AND trackings.user_uuid=users.uuid
                                                    AND users.email_verification_token IS NULL
                                                    AND users.email_notifications_enabled=1;");
        $get_recipients->bindParam(":person_name", $person_name);
        $get_recipients->execute();
        return array_column($get_recipients->fetchAll(PDO::FETCH_ASSOC), "email");
    }


    /**
     * Lists all unique names being tracked in the database.
     *
     * @return string[] all unique names in the database
     */
    public function list_all_unique_person_names(): array
    {
        $stmt = $this->conn->prepare("SELECT ALL name FROM people;");
        $stmt->execute();
        return array_column($stmt->fetchAll(PDO::FETCH_ASSOC), "name");
    }

    /**
     * Updates trackings for the given names.
     *
     * @param string[] $names the names of the trackings to update
     * @return void
     */
    public function update_trackings(array $names): void
    {
        if (empty($names)) return;

        try {
            $people_statuses = $this->mediawiki->query_page_info($names);
        } catch (Exception $exception) {
            $this->logger->error("Failed to retrieve page information.", ["cause" => $exception, "pages" => $names]);
            return;
        }

        Database::transaction($this->conn, function () use ($people_statuses) {
            $this->process_redirects($people_statuses->redirects);
            $this->process_deletions($people_statuses->missing);
            $this->process_statuses($people_statuses->results);
            return Response::satisfied();
        });
    }

    /**
     * Renames people in the database.
     *
     * @param array<string, string> $redirects a map of all changes, from old name to new name
     * @return void
     */
    private function process_redirects(array $redirects): void
    {
        // Query to rename person
        $rename = $this->conn->prepare("UPDATE people SET name=:new_name WHERE name=:old_name;");
        $rename->bindParam(":old_name", $from);
        $rename->bindParam(":new_name", $to);

        // Query to see if row with new name already exists
        $merge_needed = $this->conn->prepare("SELECT EXISTS(SELECT 1 FROM people WHERE name=:new_name);");
        $merge_needed->bindParam(":new_name", $to);

        // Queries to merge old name row with new name row
        $merge_update = $this->conn->prepare("UPDATE OR IGNORE trackings
                                              SET person_name=:new_name
                                              WHERE person_name=:old_name;");
        $merge_update->bindParam(":old_name", $from);
        $merge_update->bindParam(":new_name", $to);

        $merge_remove = $this->conn->prepare("DELETE FROM people WHERE name=:old_name;");
        $merge_remove->bindParam(":old_name", $from);

        // Perform queries
        foreach ($redirects as $from => $to) {
            if ($from === $to) continue;

            $merge_needed->execute();
            if ($merge_needed->fetch()[0] === 1) {
                $merge_update->execute();
                $merge_remove->execute();
            } else {
                $rename->execute();
            }
        }
    }

    /**
     * Deletes people from the database.
     *
     * @param string[] $deletions list of names of people to remove from the database
     * @return void
     */
    private function process_deletions(array $deletions): void
    {
        // Query to delete person, returning `name` to determine whether something changed
        $delete = $this->conn->prepare("UPDATE people
                                        SET is_deleted=1
                                        WHERE name=:name AND is_deleted<>1
                                        RETURNING name;");
        $delete->bindParam(":name", $deleted_name);

        foreach ($deletions as $deleted_name) {
            $delete->execute();
            $deleted = sizeof($delete->fetchAll(PDO::FETCH_ASSOC)) > 0;
            if ($deleted)
                foreach ($this->get_recipients($deleted_name) as $user_email)
                    $this->mailer->queue_email(new NotifyArticleDeletedEmail($user_email, $deleted_name));
        }
    }

    /**
     * Updates peoples' statuses.
     *
     * @param array<string, array{"status": \com\fwdekker\deathnotifier\trackings\PersonStatus|null, "type": ArticleType}> $statuses the current statuses of
     * people
     * @return void
     */
    private function process_statuses(array $statuses): void
    {
        // TODO: Restrict number of notifications to 1 per hour (excluding "oops we're not sure" message)
        // Query to mark person as no longer deleted, returning `name` to determine whether something changed
        $undelete = $this->conn->prepare("UPDATE people
                                          SET is_deleted=0
                                          WHERE name=:name AND is_deleted<>0
                                          RETURNING name;");
        $undelete->bindParam(":name", $person_name);

        // Query to update status, returning `name` to determine whether something changed
        $set_status = $this->conn->prepare("UPDATE people
                                            SET status=:status
                                            WHERE name=:name AND status<>:status
                                            RETURNING name;");
        $set_status->bindParam(":status", $person_status);
        $set_status->bindParam(":name", $person_name);

        foreach ($statuses as $person_name => $person_info) {
            if ($person_info["status"] === null) continue;
            $person_status = $person_info["status"]->value;

            $undelete->execute();
            $undeleted = sizeof($undelete->fetchAll(PDO::FETCH_ASSOC)) > 0;
            if ($undeleted)
                foreach ($this->get_recipients($person_name) as $user_email)
                    $this->mailer->queue_email(new NotifyArticleUndeletedEmail($user_email, $person_name));

            $set_status->execute();
            $status_changed = sizeof($set_status->fetchAll(PDO::FETCH_ASSOC)) > 0;
            if ($status_changed)
                foreach ($this->get_recipients($person_name) as $user_email)
                    $this->mailer->queue_email(new NotifyStatusChangedEmail($user_email, $person_name, $person_status));
        }
    }
}
